from typing import List, Dict, Tuple, Optional, Union
import re
from reg_automata import DFA, NFA, NFAE, State, Character, Eps
from reg_grammars import REG, Terminal
import antlr4 # type: ignore
from DFA_grammarLexer import DFA_grammarLexer
from DFA_grammarParser import DFA_grammarParser
from DFA_grammarListener import DFA_grammarListener

# TODO ask: commas at the ends of lines or not?
class Parser:
    def __init__(self):
        pass

    def reg_to_str(self, reg: REG, full: bool = False) -> str:
        rules = self.rules_to_str(reg.rules)
        if not full:
            return "P=" + rules

        # full - verbose description of DFA - only for development, dismiss later
        nonterminals = "{" + ','.join(set(map(lambda x: x.name, reg.nonterminals))) + "}"
        terminals = "{" + ','.join(set(map(lambda x: x.name, reg.terminals))) + "}"

        out = "REG = (" + nonterminals + ", " + terminals + ", P," + reg.init.name + ")\nP="
        out += (self.rules_to_str(reg.rules))
        return out

    def rules_to_str(self, rules: REG.Rules) -> str:
        out = ""
        for rule in rules:
            out = out[:-1] + (",\n")
            rewrite = rules[rule]
            if len(rewrite) == 0:
                continue

            out += (rule.name + "->")
            for variant in rewrite:
                if isinstance(variant, Terminal):
                    out += variant.name
                else:
                    for i in range(len(variant)):
                        out += variant[i].name
                if next:  # TODO
                    out += "|"

        return "{" + out[2:len(out) - 1] + "}"

    #    def str_to_reg(self, source: str) -> REG:
    #        re.match(...((nt),(t),P,S)\n rest

    def dfa_to_str(self, dfa: DFA, full : bool = False) -> str:

        # full - verbose description of DFA - only for development, dismiss later
        states = "{" + ','.join(set(map(lambda x: x.name, dfa.states))) + "}"
        characters = "{" + ','.join(set(map(lambda x: x.name, dfa.characters))) + "}"

        transition = ""
        for key, state_2 in dfa.transition.items():
            state_1, character = key
            transition += "(" + state_1.name + "," + character.name + ")=" + state_2.name + " "

        init = "init=" + dfa.init.name
        final = "final={" + ','.join(set(map(lambda x: x.name, dfa.final))) + "}"

        # full - verbose description of DFA - only for development, dismiss later
        if full:
            return "DFA = (" + states + ", " + characters + ", d, " + \
                   init + ", " + final + ")\n" + transition

        else:
            return init + " " + transition + final


    def nfa_to_str(self, nfa: Union[NFA, NFAE], full : bool = False) -> str:

        # full - verbose description of DFA - only for development, dismiss later
        states = "{" + ','.join(set(map(lambda x: x.name, nfa.states))) + "}"
        characters = "{" + ','.join(set(map(lambda x: x.name, nfa.characters))) + "}"

        transition = ""
        for key, set_states in nfa.transition.items():
            state, character = key
            char = character.name if not isinstance(character, Eps) else "ε"
            if isinstance(character, Eps) and isinstance(nfa, NFAE):
                dest_states = nfa.transition[state, character]
            elif isinstance(character, Character) and isinstance(nfa, NFA):
                dest_states = nfa.transition[state, character]
            transition += "(" + state.name + "," + char + ")={" + \
                          ','.join(set(map(lambda x: x.name, dest_states))) + "}\n"

        init = "init=" + nfa.init.name
        final = "final={" + ','.join(set(map(lambda x: x.name, nfa.final))) + "}"

        if full:
            return "NFA = (" + states + ", " + characters + ", d, " + \
                   init + ", " + final + ")\n" + transition

        else:
            return transition + "\n" + final


    def str_to_dfa(self, input: str):
        input = "init=0 (0,x)=1 (0,y)=2 (1,x)=1 (1,y)=1 (2,x)=0 (2,y)=4 (3,x)=2 (3,y)=5 (4,x)=3 (4,y)=0 (5,x)=5 (5,y)=5 final={4,5}"
        chars = antlr4.InputStream(input)
        lexer = DFA_grammarLexer(chars)
        tokens = antlr4.CommonTokenStream(lexer)
        parser = DFA_grammarParser(tokens)

        tree = parser.start()
        printer = StatenamePrinter()
        walker = antlr4.ParseTreeWalker()
        walker.walk(printer, tree)
        dfa = DFA(printer.states, printer.characters, printer.transition, printer.init, printer.final)
        return dfa

class StatenamePrinter(DFA_grammarListener):
    def __init__(self):
        self.states = set()
        self.characters = set()
        self.transition = {}
        self.init = State("")
        self.final = set()

    def exitInit(self, ctx):
        state = State(str(ctx.statename().STATE()))
        self.init = state
        self.states.add(state)

    def exitProduction(self, ctx):
        state = State(str(ctx.statename(0).STATE()))
        character = Character(str(ctx.statename(1).STATE()))
        dest_state = State(str(ctx.statename(2).STATE()))
        self.states.add(state)
        self.states.add(dest_state)
        self.characters.add(character)
        self.transition[state, character] = dest_state

    def exitFinal(self, ctx):
        i = 0
        while ctx.statename(i) is not None:
            state = State(str(ctx.statename(i).STATE()))
            self.states.add(state)
            self.final.add(state)
            i += 1
